Attribute VB_Name = "modShellAndWaitEnhancedVB6"
Option Explicit
Option Compare Text
Option Base 1
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' modShellAndWaitEnhancedVB6
' By Chip Pearson, chip@cpearson.com, www.cpearson.com
' See www.cpearson.com/excel/shellandwait.htm for details and documentations.
'
' This module contains the F_7_AB_1_ShellAndWaitEnhanced function and all supporting constants, declarations, and
' variables. This module is completely self-contained. It requires no modification or additional code.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Constants
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Private Const C_7_AB_1_DOEVENTS_TIMEOUT_MILLISECONDS As Long = 500&
Private Const C_7_AB_1_VBA_FORM_CLASSNAME = "ThunderDFrame"
Private Const C_7_AB_1_VB6_FORM_CLASSNAME = "ThunderFormDC"

''''''''''''''''''''''''''''''''
' Error Numbers
''''''''''''''''''''''''''''''''
Public Const C_7_AB_1_ERR_ENHANCED_NO_ERROR As Long = 0
Public Const C_7_AB_1_ERR_ENHANCED_BLANK_COMMAND_LINE = 1
Public Const C_7_AB_1_ERR_ENHANCED_SHELL_ALREADY_RUNNING = 2
Public Const C_7_AB_1_ERR_ENHANCED_BOTH_HWND_AND_CONTROL_ARE_EMPTY = 3
Public Const C_7_AB_1_ERR_ENHANCED_BOTH_HWND_AND_CONTROL_ARE_NOT_EMPTY = 4
Public Const C_7_AB_1_ERR_ENHANCED_HWND_IS_NOT_A_VALID_WINDOW = 5
Public Const C_7_AB_1_ERR_ENHANCED_HWND_IS_NOT_A_VISIBLE_WINDOW = 6
Public Const C_7_AB_1_ERR_ENHANCED_CONTROL_DOES_NOT_SUPPORT_A_TEXT_OR_CAPTION_PROPERTY = 7
Public Const C_7_AB_1_ERR_ENHANCED_ERROR_WITH_SETWINDOWTEXT = 8
Public Const C_7_AB_1_ERR_ENHANCED_ARGUMENT_IS_NOT_ARRAY = 9
Public Const C_7_AB_1_ERR_ENHANCED_ARRAY_IS_NOT_ALLOCATED = 10
Public Const C_7_AB_1_ERR_ENHANCED_ARRAY_ELEMENT_IS_NOT_STRING = 11
Public Const C_7_AB_1_ERR_ENHANCED_ERROR_FROM_SHELL = 12
Public Const C_7_AB_1_ERR_ENHANCED_ERROR_FROM_OPEN_PROCESS = 13
Public Const C_7_AB_1_ERR_ENHANCED_WAIT_ABANDONED = 14
Public Const C_7_AB_1_ERR_ENHANCED_WAIT_FAILED = 15
Public Const C_7_AB_1_ERR_EHNANCED_UNEXPECTED_ERROR = 16
Public Const C_7_AB_1_ERR_ENHANCED_ERROR_WITH_SETTIMER = 17
''''''''''''''''''''''''''''''''
' CONSTANTS USED WITH VARIOUS API
' PROCEDURES
''''''''''''''''''''''''''''''''
Private Const WAIT_INFINITE = -1&
Private Const SYNCHRONIZE = &H100000
Private Const STATUS_ABANDONED_WAIT_0 As Long = &H80
Private Const STATUS_WAIT_0 As Long = &H0
Private Const WAIT_ABANDONED As Long = (STATUS_ABANDONED_WAIT_0 + 0)
Private Const WAIT_OBJECT_0 As Long = (STATUS_WAIT_0 + 0)
Private Const WAIT_TIMEOUT As Long = &H102
Private Const LB_SETTABSTOPS As Long = &H192
Private Const WAIT_FAILED As Long = &HFFFFFFFF '  = -1
Private Const STATUS_USER_APC As Long = &HC0
Private Const WAIT_IO_COMPLETION As Long = STATUS_USER_APC
Private Const STATUS_TIMEOUT As Long = 258&
Private Const C_7_AB_1_FORMAT_MESSAGE_TEXT_LEN = 160

'''''''''''''''''''''''''''''''''
' Used with GetAncestor
'''''''''''''''''''''''''''''''''
Private Const GA_PARENT As Long = 1
Private Const GA_ROOT As Long = 2
Private Const GA_ROOTOWNER As Long = 3

''''''''''''''''''''''''''''''''
' Used by FormatMessage
''''''''''''''''''''''''''''''''
Private Const FORMAT_MESSAGE_ALLOCATE_BUFFER = &H100
Private Const FORMAT_MESSAGE_ARGUMENT_ARRAY = &H2000
Private Const FORMAT_MESSAGE_FROM_HMODULE = &H800
Private Const FORMAT_MESSAGE_FROM_STRING = &H400
Private Const FORMAT_MESSAGE_FROM_SYSTEM = &H1000
Private Const FORMAT_MESSAGE_IGNORE_INSERTS = &H200
Private Const FORMAT_MESSAGE_MAX_WIDTH_MASK = &HFF
            
Private Const STANDARD_RIGHTS_ALL As Long = &H1F0000
Private Const STANDARD_RIGHTS_REQUIRED As Long = &HF0000
Private Const PROCESS_ALL_ACCESS As Long = (STANDARD_RIGHTS_REQUIRED Or SYNCHRONIZE Or &HFFF)


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Windows API Declares
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

Private Declare Function SetTimer Lib "user32" ( _
    ByVal hwnd As Long, _
    ByVal nIDEvent As Long, _
    ByVal uElapse As Long, _
    ByVal lpTimerFunc As Long) As Long

Private Declare Function KillTimer Lib "user32" ( _
    ByVal hwnd As Long, _
    ByVal nIDEvent As Long) As Long

Private Declare Function CloseHandle Lib "kernel32" ( _
    ByVal hObject As Long) As Long
            
Private Declare Function OpenProcess Lib "kernel32.dll" ( _
    ByVal dwDesiredAccess As Long, _
    ByVal bInheritHandle As Long, _
    ByVal dwProcessId As Long) As Long

Private Declare Function WaitForSingleObject Lib "kernel32.dll" ( _
    ByVal hHandle As Long, _
    ByVal dwMilliseconds As Long) As Long
            
Private Declare Function FindWindow Lib "user32" Alias "FindWindowA" ( _
    ByVal lpClassName As String, _
    ByVal lpWindowName As String) As Long

Private Declare Function FormatMessage Lib "kernel32" Alias "FormatMessageA" ( _
    ByVal dwFlags As Long, _
    lpSource As Any, _
    ByVal dwMessageId As Long, _
    ByVal dwLanguageId As Long, _
    ByVal lpBuffer As String, _
    ByVal nSize As Long, Arguments As Long) As Long

Private Declare Function IsWindow Lib "user32" ( _
    ByVal hwnd As Long) As Long

Private Declare Function IsWindowVisible Lib "user32" ( _
    ByVal hwnd As Long) As Long

Private Declare Function SetWindowText Lib "user32" Alias "SetWindowTextA" ( _
    ByVal hwnd As Long, _
    ByVal lpString As String) As Long

Private Declare Function GetAncestor Lib "user32.dll" ( _
    ByVal hwnd As Long, _
    ByVal gaFlags As Long) As Long


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Private Variables
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Private P_7_AB_1_MsgNdx As Long
Private P_7_AB_1_PropertyName As String
Private P_7_AB_1_MessageArray As Variant
Private P_7_AB_1_MessageControl As Object
Private P_7_AB_1_TimerID As Long
Private P_7_AB_1_HWnd As Long
Private P_7_AB_1_MessageHWnd As Long
Private P_7_AB_1_FormHWnd As Long
Private P_7_AB_1_ShellStillRunning As Boolean
Private P_7_AB_1_Abort As Boolean
Private P_7_AB_1_CommandLine As String



Public Function F_7_AB_1_ShellAndWaitEnhanced(A_7_AB_1_CommandLine As String, _
            A_7_AB_1_MessageControl As VB.Control, _
            A_7_AB_1_MessageHWnd As Long, _
            A_7_AB_1_MessageCycleTimeSeconds As Long, _
            A_7_AB_1_MessageArray As Variant, _
            A_7_AB_1_DoEventsMilliseconds As Long, _
            ByRef A_7_AB_1_ErrorNumber As Long, _
            ByRef A_7_AB_1_ErrorText As String, _
            Optional A_7_AB_1_MessageOnClose As String = vbNullString) As Boolean
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' F_7_AB_1_ShellAndWaitEnhanced
' Shell's out to A_7_AB_1_CommandLine.
' Parameter                         Description
' ----------------------------      ------------------------------------------------------------------------
' A_7_AB_1_CommandLine              The command line that is passed to the Shell commamnd.
'
' A_7_AB_1_MessageControl           This is a reference to a control on a UserForm in which the wait
'                                   messages will be displayed. This control must have a "Caption" or
'                                   "Text" property. If this paramter is not NOTHING, A_7_AB_1_MessageHwnd
'                                   must be 0. An error will occur if A_7_AB_1_MessageControl is not
'                                   Nothing and A_7_AB_1_MessageHWnd is not 0. In short you can identify
'                                   the control to display the messages either as a reference to the
'                                   control or by its HWnd, but not both.
'
' A_7_AB_1_MessageHWnd              This is the HWnd of the control that is to display the messages.
'                                   If A_7_AB_1_MessageHWnd is not 0, A_7_AB_1_MessageControl must
'                                   be NOTHING. A_7_AB_1_MessageHWnd must the the HWnd of a control
'                                   whose text can be set with SetWindowText. An error will occur
'                                   if the control does not support text update via SetWindowText.
'
' A_7_AB_1_MessageCycleTimeSeconds  This value indicates the number of seconds each message in
'                                   A_7_AB_1_MessageArray should be displayed on A_7_AB_1_MessageControl.
'
' A_7_AB_1_MessageArray             This is an array of "please wait" type strings that are sequentially
'                                   displayed on A_7_AB_1_MessageControl. Each string in the array is
'                                   displayed for A_7_AB_1_MessageCycleTimeSeconds seconds.
'                                   The messages repeat in a loop as long as the Shell'd program is running.
'
' A_7_AB_1_DoEventsMilliseconds     This value is the number of milliseconds that we will wait before
'                                   calling DoEvents to allow other threads to execute. Set this value
'                                   to 0 to use the default value. This timer is independent of the
'                                   A_7_AB_1_MessageCycleTimeSeconds value. A typical value for this
'                                   parameter is 500.
'
' A_7_AB_1_ErrorNumber              is a reference to a Long type variable that will receive the
'                                   procedure error value is an error occured.
'
' A_7_AB_1_ErrorText                is a reference to a String type variable that will receive the
'                                   descriptive error text associated with A_7_AB_1_ErrorNumber.

' A_7_AB_1_MessageOnClose           This is the message to be displayed when the shell'd program
'                                   ends. This is optional. If omitted or vbNullString, no close message
'                                   is displayed. This message is the last message displayed by this
'                                   procedure. It is up to calling procedure to clear the text
'                                   when this procedure returns.
'
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

Dim V_7_AB_1_DoEventsTimeOutMilliseconds  As Long
Dim V_7_AB_1_TaskID As Double
Dim V_7_AB_1_WaitRes As Long
Dim V_7_AB_1_ProcessHandle As Long
Dim V_7_AB_1_ErrorNumber As Long
Dim V_7_AB_1_ErrorText As String
Dim V_7_AB_1_Res As Long

''''''''''''''''''''''''''''''''''''''''''''
' Set the passed in error variables to the
' default "no error" values
''''''''''''''''''''''''''''''''''''''''''''
A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_NO_ERROR
A_7_AB_1_ErrorText = vbNullString

'''''''''''''''''''''''''''''''''''''''''''''
' Initialize the return code to False. It will
' be set to True when we successfully exit the
' wait loop.
''''''''''''''''''''''''''''''''''''''''''''''
F_7_AB_1_ShellAndWaitEnhanced = False


''''''''''''''''''''''''''''''''''''''''''''''''''
' ShellAndWaitEnhanced allows for only a single
' Shell'd program at one time.
'''''''''''''''''''''''''''''''''''''''''''''''''''
If P_7_AB_1_ShellStillRunning = True Then
    A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_SHELL_ALREADY_RUNNING
    A_7_AB_1_ErrorText = "ShellAndWaitEnhanced is already running one Shell'd program. It supports only one program at a time."
    Exit Function
End If


'''''''''''''''''''''''''''''''''''''''''''''
' ensure that we have a valid command line.
''''''''''''''''''''''''''''''''''''''''''''''
If Trim(A_7_AB_1_CommandLine) = vbNullString Then
    A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_BLANK_COMMAND_LINE
    A_7_AB_1_ErrorText = "The command line to F_7_AB_1_ShellAndWaitEnhanced was blank."
    F_7_AB_1_ShellAndWaitEnhanced = False
    Exit Function
End If

''''''''''''''''''''''''''''''''''''''''''
' Save a copy of the command line
''''''''''''''''''''''''''''''''''''''''''
P_7_AB_1_CommandLine = A_7_AB_1_CommandLine



'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' ensure that
' A_7_AB_1_MessageControl is NOTHING and A_7_AB_1_MessageHWnd is 0, or
' A_7_AB_1_MessageControl is not NOTHING and A_7_AB_1_MessageHWnd <> 0.
' Display an error if both A_7_AB_1_MessageControl is NOTHING and
' A_7_AB_1_MessageHWnd is 0, or if both A_7_AB_1_MessageControl is Not NOTHING
' and A_7_AB_1_MessageHWnd is <> 0. Either, but not both, A_7_AB_1_MessageControl or
' A_7_AB_1_MessageHWnd must be specified as the destination for the messages.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

If A_7_AB_1_MessageHWnd = 0 Then
    If A_7_AB_1_MessageControl Is Nothing Then
        A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_BOTH_HWND_AND_CONTROL_ARE_EMPTY
        A_7_AB_1_ErrorText = "The Message HWnd is 0 and the Message Control is Nothing. One of these must be set to receive messages."
        F_7_AB_1_ShellAndWaitEnhanced = False
        Exit Function
    Else
        Set P_7_AB_1_MessageControl = A_7_AB_1_MessageControl
        P_7_AB_1_FormHWnd = F_7_AB_1_GetHWndOfFormByControlVB6(A_7_AB_1_FormControl:=A_7_AB_1_MessageControl)
        P_7_AB_1_MessageHWnd = 0&
    End If
Else
    If Not A_7_AB_1_MessageControl Is Nothing Then
        A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_BOTH_HWND_AND_CONTROL_ARE_NOT_EMPTY
        A_7_AB_1_ErrorText = "The Message HWnd was not zero and the Message Control is not Nothing. Only one of these parameters may be used."
        F_7_AB_1_ShellAndWaitEnhanced = False
        Exit Function
    Else
        ''''''''''''''''''''''''''''''''''''''''''
        ' ensure we have a valid window
        ''''''''''''''''''''''''''''''''''''''''''
        If IsWindow(A_7_AB_1_MessageHWnd) = False Then
            A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_HWND_IS_NOT_A_VALID_WINDOW
            A_7_AB_1_ErrorText = "The HWnd in A_7_AB_1_MessageHWnd passed to F_7_AB_1_ShellAndWaitEnhanced is not a valid window."
            F_7_AB_1_ShellAndWaitEnhanced = False
            Exit Function
        End If
        If IsWindowVisible(A_7_AB_1_MessageHWnd) = False Then
            A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_HWND_IS_NOT_A_VISIBLE_WINDOW
            A_7_AB_1_ErrorText = "The HWnd in A_7_AB_1_MessageHWnd passed to F_7_AB_1_ShellAndWaitEnhanced is not a visible window."
            F_7_AB_1_ShellAndWaitEnhanced = False
            Exit Function
        End If
        '''''''''''''''''''''''''''''''''''''''''''''''''''''''''
        ' if A_7_AB_1_MessageHWnd passed the valid windows tests,
        ' set our module level variable to the HWnd.
        '''''''''''''''''''''''''''''''''''''''''''''''''''''''''
        P_7_AB_1_MessageHWnd = A_7_AB_1_MessageHWnd
    End If
End If
    
On Error Resume Next
Err.Clear
'''''''''''''''''''''''''''''''''''''''''''''
' Use CallByName to set the text of
' A_7_AB_1_MessageControl's Text or Caption
' property. Store the property name in
' P_7_AB_1_PropertyName for future use. If
' A_7_AB_1_MessageControl has neither a "Text"
' or "Caption" property, the function exits
' with a result of False. The wait display messages
' in A_7_AB_1_MessageArray are displayed in
' the Text or Caption property of A_7_AB_1_MessageControl.
'''''''''''''''''''''''''''''''''''''''''''''
If P_7_AB_1_MessageHWnd = 0 Then
    ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
    ' If P_7_AB_1_MessageHWnd is 0, we were passed a reference to a form
    ' control in which to display the messages. Ensure that that control
    ' supports either the "Caption" or "Text" property.
    ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
    CallByName A_7_AB_1_MessageControl, "Caption", VbLet, "Initializing"
    If Err.Number = 0 Then
        P_7_AB_1_PropertyName = "Caption"
    Else
        Err.Clear
        CallByName A_7_AB_1_MessageControl, "Text", VbLet, "Initializing"
        If Err.Number = 0 Then
            P_7_AB_1_PropertyName = "Text"
        Else
            A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_CONTROL_DOES_NOT_SUPPORT_A_TEXT_OR_CAPTION_PROPERTY
            A_7_AB_1_ErrorText = "The control A_7_AB_1_MessageControl does not have a Caption or Text property."
            F_7_AB_1_ShellAndWaitEnhanced = False
            Exit Function
        End If
    End If

Else
    ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
    ' Since P_7_AB_1_MessageHWnd is not 0, we were passed an HWnd in which
    ' to display the messages. Ensure we can use SetWindowText with this
    ' HWnd.
    ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
    V_7_AB_1_Res = SetWindowText(P_7_AB_1_MessageHWnd, "Initializing")
    If V_7_AB_1_Res = 0& Then
        V_7_AB_1_ErrorNumber = Err.LastDllError
        V_7_AB_1_ErrorText = F_7_AB_1_GetSystemErrorMessageText(A_7_AB_1_ErrorNumber:=V_7_AB_1_ErrorNumber)
        A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_ERROR_WITH_SETWINDOWTEXT
        A_7_AB_1_ErrorText = "An error occurred in SetWindowText with HWnd: " & CStr(P_7_AB_1_HWnd) & vbCrLf & _
            "System Error: " & CStr(V_7_AB_1_ErrorNumber) & vbCrLf & _
            "Description:  " & V_7_AB_1_ErrorText
        F_7_AB_1_ShellAndWaitEnhanced = False
        Exit Function
    End If

End If
On Error GoTo 0

''''''''''''''''''''''''''''''''''''''''''''''
' ensure that A_7_AB_1_MessageCycleTimeSeconds
' is at least one second.
''''''''''''''''''''''''''''''''''''''''''''''
If A_7_AB_1_MessageCycleTimeSeconds < 1 Then
    MsgBox "The value in A_7_AB_1_MessageCycleTimeSeconds is less than 1 second."
    F_7_AB_1_ShellAndWaitEnhanced = False
    Exit Function
End If

''''''''''''''''''''''''''''''''''''''''''''''
' ensure that we have an array of strings in
' A_7_AB_1_MessageArray
''''''''''''''''''''''''''''''''''''''''''''''
If IsArray(A_7_AB_1_MessageArray) = False Then
    A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_ARGUMENT_IS_NOT_ARRAY
    A_7_AB_1_ErrorText = "The parameter A_7_AB_1_MessageArray to F_7_AB_1_ShellAndWaitEnhanced is not an array." & vbCrLf & _
        "Its TypeName is: '" & TypeName(A_7_AB_1_MessageArray) & "'."
    F_7_AB_1_ShellAndWaitEnhanced = False
    Exit Function
End If

If F_7_AB_1_IsArrayAllocated(A_7_AB_1_MessageArray) = False Then
    A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_ARRAY_IS_NOT_ALLOCATED
    A_7_AB_1_ErrorText = "The array A_7_AB_1_MessageArray is an unallocated array."
    F_7_AB_1_ShellAndWaitEnhanced = False
    Exit Function
End If

If VarType(A_7_AB_1_MessageArray(LBound(A_7_AB_1_MessageArray))) <> vbString Then
    A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_ARRAY_ELEMENT_IS_NOT_STRING
    A_7_AB_1_ErrorNumber = "The array A_7_AB_1_MessageArray is an does not contains String type values."
    F_7_AB_1_ShellAndWaitEnhanced = False
    Exit Function
End If
   
'''''''''''''''''''''''''''''''''''''''''''''''''
' If we passed the array tests, set our module
' level variable to the message array.
'''''''''''''''''''''''''''''''''''''''''''''''''
P_7_AB_1_MessageArray = A_7_AB_1_MessageArray
      
''''''''''''''''''''''''''''''''''''''''''''''
' Test the value of A_7_AB_1_DoEventsMilliseconds
' to see if we should use the supplied value or
' the default value.
''''''''''''''''''''''''''''''''''''''''''''''
If A_7_AB_1_DoEventsMilliseconds <= 0 Then
    V_7_AB_1_DoEventsTimeOutMilliseconds = C_7_AB_1_DOEVENTS_TIMEOUT_MILLISECONDS
Else
    V_7_AB_1_DoEventsTimeOutMilliseconds = A_7_AB_1_DoEventsMilliseconds
End If
    
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the Windows Timer to A_7_AB_1_MessageCycleTimeSeconds * 1000
' milliseconds. Windows will automatically call S_7_AB_1_TimerProc
' every time the timer "pops". We don't use an HWnd with the timer,
' so that we can still kill the timer even if the user closes
' the message form while the shelled process is still running.
' In practice, the userform's QueryClose event should prevent the
' user from closing the form while the shell'd process is still
' running.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
P_7_AB_1_TimerID = SetTimer(hwnd:=0&, nIDEvent:=0&, _
        uElapse:=CLng(A_7_AB_1_MessageCycleTimeSeconds * 1000), lpTimerFunc:=AddressOf S_7_AB_1_TimerProc)
If P_7_AB_1_TimerID = 0 Then
    ' an error occurred
    V_7_AB_1_ErrorNumber = Err.LastDllError
    V_7_AB_1_ErrorText = F_7_AB_1_GetSystemErrorMessageText(A_7_AB_1_ErrorNumber:=V_7_AB_1_ErrorNumber)
    A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_ERROR_WITH_SETTIMER
    A_7_AB_1_ErrorText = "An error occurred setting the Windows Timer with SetTimer." & vbCrLf & _
        "System Error Number: " & CStr(V_7_AB_1_ErrorNumber) & vbCrLf & _
        "Description:         " & V_7_AB_1_ErrorText
    F_7_AB_1_ShellAndWaitEnhanced = False
    Exit Function
End If

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' This is the actual call to Shell.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
On Error Resume Next

P_7_AB_1_ShellStillRunning = True
Err.Clear
V_7_AB_1_TaskID = Shell(A_7_AB_1_CommandLine, vbNormalFocus)
If Err.Number <> 0 Then
    '''''''''''''''''''''''''''''''''''''''''
    ' an error occure with the SHELL command.
    '''''''''''''''''''''''''''''''''''''''''
    V_7_AB_1_ErrorNumber = Err.LastDllError
    A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_ERROR_FROM_SHELL
    V_7_AB_1_ErrorText = F_7_AB_1_GetSystemErrorMessageText(A_7_AB_1_ErrorNumber:=V_7_AB_1_ErrorNumber)
    A_7_AB_1_ErrorText = "The Shell command failed with command line """ & A_7_AB_1_CommandLine & """" & vbCrLf & _
        "VBA Error:         " & CStr(Err.Number) & vbCrLf & _
        "Error Description: " & Err.Description & vbCrLf & _
        "System Error:      " & CStr(V_7_AB_1_ErrorNumber) & vbCrLf & _
        "Error Description: " & V_7_AB_1_ErrorText
    
    S_7_AB_1_KillTheTimer
    P_7_AB_1_ShellStillRunning = False
    F_7_AB_1_ShellAndWaitEnhanced = False
    Exit Function
End If
On Error GoTo 0

V_7_AB_1_ProcessHandle = OpenProcess(SYNCHRONIZE, False, V_7_AB_1_TaskID)
If V_7_AB_1_ProcessHandle = 0& Then
    V_7_AB_1_ErrorNumber = Err.LastDllError
    V_7_AB_1_ErrorText = F_7_AB_1_GetSystemErrorMessageText(A_7_AB_1_ErrorNumber:=V_7_AB_1_ErrorNumber)
    A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_ERROR_FROM_OPEN_PROCESS
    A_7_AB_1_ErrorText = "An error occurred with OpenProcess." & vbCrLf & _
        "System Error:      " & CStr(V_7_AB_1_ErrorNumber) & vbCrLf & _
        "Error Description: " & V_7_AB_1_ErrorText
    S_7_AB_1_KillTheTimer
    F_7_AB_1_ShellAndWaitEnhanced = False
    Exit Function
End If

'''''''''''''''''''''''''''''''''''''''''''''''''''
' Call WaitForSingleObject. It will return every
' V_7_AB_1_DoEventsTimeOutMilliseconds with an
' WAIT_TIMEOUT result. In this case, we call
' DoEvents to let other things happen.
'''''''''''''''''''''''''''''''''''''''''''''''''''
V_7_AB_1_WaitRes = WaitForSingleObject(V_7_AB_1_ProcessHandle, V_7_AB_1_DoEventsTimeOutMilliseconds)

''''''''''''''''''''''''''''''''''''''''''''''''''''
' We loop until the result of WaitForSingleObject is WAIT_OBJECT_0,
' which indicates that the shell'd program has ended
''''''''''''''''''''''''''''''''''''''''''''''''''''
Do Until (V_7_AB_1_WaitRes = WAIT_OBJECT_0) Or (P_7_AB_1_Abort = True)
    ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
    ' Here, we loop, timing out at V_7_AB_1_DoEventsTimeOutMilliseconds and calling DoEvents. We'll
    ' exit the loop when V_7_AB_1_Res is any value other than WAIT_TIMEOUT
    ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
    V_7_AB_1_WaitRes = WaitForSingleObject(V_7_AB_1_ProcessHandle, V_7_AB_1_DoEventsTimeOutMilliseconds)
    DoEvents
    Select Case V_7_AB_1_WaitRes
        Case WAIT_ABANDONED
            V_7_AB_1_ErrorNumber = Err.LastDllError
            V_7_AB_1_ErrorText = F_7_AB_1_GetSystemErrorMessageText(V_7_AB_1_ErrorNumber)
            A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_WAIT_ABANDONED
            A_7_AB_1_ErrorText = "An error occurred with the WaitForSingleObject function call: WAIT_ABANDONED." & vbCrLf & _
                    "System Error Number: " & CStr(V_7_AB_1_ErrorNumber) & vbCrLf & _
                    "Error Description:   " & F_7_AB_1_GetSystemErrorMessageText(V_7_AB_1_ErrorNumber)
            F_7_AB_1_ShellAndWaitEnhanced = False
            Exit Do
        Case WAIT_OBJECT_0
            ' the shell'd process closed
            F_7_AB_1_ShellAndWaitEnhanced = True
            Exit Do
        Case WAIT_FAILED
            V_7_AB_1_ErrorNumber = Err.LastDllError
            V_7_AB_1_ErrorText = F_7_AB_1_GetSystemErrorMessageText(V_7_AB_1_ErrorNumber)
            A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_ENHANCED_WAIT_FAILED
            A_7_AB_1_ErrorText = "An error occurred with the WaitForSingleObject function call: WAIT_FAILED" & vbCrLf & _
                    "System Error Number: " & CStr(V_7_AB_1_ErrorNumber) & vbCrLf & _
                    "Error Description:   " & F_7_AB_1_GetSystemErrorMessageText(V_7_AB_1_ErrorNumber)
            F_7_AB_1_ShellAndWaitEnhanced = False
            Exit Do
        Case WAIT_TIMEOUT
            ' this is when the loop times out at C_7_AB_1_DOEVENTS_INTERVAL_MILLISECONDS.
            ' do nothing
            
        Case Else
            V_7_AB_1_ErrorNumber = Err.LastDllError
            V_7_AB_1_ErrorText = F_7_AB_1_GetSystemErrorMessageText(V_7_AB_1_ErrorNumber)
            A_7_AB_1_ErrorNumber = C_7_AB_1_ERR_EHNANCED_UNEXPECTED_ERROR
            A_7_AB_1_ErrorText = "An unexpected error occurred with the WaitForSingleObject function call." & vbCrLf & _
                    "System Error Number: " & CStr(V_7_AB_1_ErrorNumber) & vbCrLf & _
                    "Error Description:   " & F_7_AB_1_GetSystemErrorMessageText(V_7_AB_1_ErrorNumber)
            F_7_AB_1_ShellAndWaitEnhanced = False
            Exit Do
    End Select
    
''''''''''''''''''''''
' END OF MAIN TIMER LOOP
''''''''''''''''''''''
Loop

P_7_AB_1_ShellStillRunning = False

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Display the Close message. Don't display if the process
' was aborted.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
If P_7_AB_1_Abort = False Then
    If A_7_AB_1_MessageOnClose <> vbNullString Then
        If P_7_AB_1_MessageHWnd = 0 Then
            CallByName P_7_AB_1_MessageControl, P_7_AB_1_PropertyName, VbLet, A_7_AB_1_MessageOnClose
        Else
            SetWindowText P_7_AB_1_MessageHWnd, A_7_AB_1_MessageOnClose
        End If
    End If
End If

''''''''''''''''''''''''''''''''''''''''
' close the process and kill the timer, and return.
' The return code (true or false) is set in the
' main loop. It gets set to false if an error
' occurred or true if we get out on a
' WAIT_OBJECT_0 result from
' WaitForSingleObject.
''''''''''''''''''''''''''''''''''''''''
CloseHandle V_7_AB_1_ProcessHandle
S_7_AB_1_KillTheTimer
P_7_AB_1_Abort = False
P_7_AB_1_CommandLine = vbNullString


End Function

Public Function F_7_AB_1_AbortShellAndWaitEnhanced()
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' F_7_AB_1_AbortShellAndWait
' This procedure should be called from the form on which messages are displayed when the form is closed and
' the shell programming is still running. It will terminate the WaitForSingleObject loop, close the process
' handle, and Kill the timer. It will leave the shell'd program running.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
    P_7_AB_1_Abort = True
End Function

Public Function F_7_AB_1_ShellEhancedStillRunning() As Boolean
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' F_7_AB_1_ShellStillRunning
' This returns TRUE or FALSE indicating whether the Shell'd program is still running.
''''''''''''''F_7_AB_1_ShellEnhancedStillRunning =''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
     F_7_AB_1_ShellEhancedStillRunning = P_7_AB_1_ShellStillRunning
End Function

Public Function F_7_AB_1_GetEnhancedCommmandLine() As String
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' F_7_AB_1_GetCommmandLine
' This function returns the command line passed to the Shell function.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
    F_7_AB_1_GetEnhancedCommmandLine = P_7_AB_1_CommandLine
End Function



Private Function F_7_AB_1_IsArrayAllocated(A_7_AB_1_Arr As Variant) As Boolean
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' F_7_AB_1_IsArrayAllocated
' Returns TRUE if the array is allocated (either a static array or a dynamic array that has been
' sized with Redim) or FALSE if the array has not been allocated (a dynamic that has not yet
' been sized with Redim, or a dynamic array that has been Erased).
'
' This function is just the reverse of F_7_AB_1_IsArrayEmpty.
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

Dim V_7_AB_1_N As Long
On Error Resume Next

' if A_7_AB_1_Arr is not an array, return FALSE and get out.
If IsArray(A_7_AB_1_Arr) = False Then
    F_7_AB_1_IsArrayAllocated = False
    Exit Function
End If

' Try to get the UBound of the array. If the array has not been allocated,
' an error will occur. Test Err.Number to see if an error occured.
V_7_AB_1_N = UBound(A_7_AB_1_Arr, 1)
If Err.Number = 0 Then
    ' no error. array has been allocated.
    F_7_AB_1_IsArrayAllocated = True
Else
    ' error. unallocated array
    F_7_AB_1_IsArrayAllocated = False
End If

End Function

Private Sub S_7_AB_1_TimerProc(hwnd As Long, uMsg As Long, idEvent As Long, dwTime As Long)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' S_7_AB_1_TimerProc
' This procedure is called by Windows every time the message timer "pops". This is the message
' timer, whose interval is set with A_7_AB_1_MessageCycleTimeSeconds.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Debug.Print "Timer Tick"

Dim V_7_AB_1_Text As String             ' This gets set to the text to display in the message control
                                        ' which is Text or Caption property set via CallByName, or via
                                        ' SetWindowText is we were passed an HWnd to
Dim V_7_AB_1_Res As Long                ' General purpose return code variable
Dim V_7_AB_1_ErrorNumber As Long        ' Error number. gets set to Err.LastDllError after API calls.
Dim V_7_AB_1_ErrorText As String        ' The text describing V_7_AB_1_ErrorNumber, retrieived by
                                        ' F_7_AB_1_GetSystemErrorMessageText


P_7_AB_1_MsgNdx = P_7_AB_1_MsgNdx + 1
' if the P_7_AB_1_MsgNdx is past the end of P_7_AB_1_MessageArray, loop back to
' the beginning of the array.
If P_7_AB_1_MsgNdx > UBound(P_7_AB_1_MessageArray) Then
    P_7_AB_1_MsgNdx = LBound(P_7_AB_1_MessageArray)
End If

' get the text out of the array
V_7_AB_1_Text = P_7_AB_1_MessageArray(P_7_AB_1_MsgNdx)

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' test to ensure we have a HWnd to a valid window (i.e., the form is still displayed)
' before trying to write out a message.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
If P_7_AB_1_MessageHWnd = 0 Then
    If IsWindow(P_7_AB_1_FormHWnd) Then
        If IsWindowVisible(P_7_AB_1_FormHWnd) Then
            ' set the P_7_AB_1_MessageControl's Text or Caption property to the current message
            CallByName P_7_AB_1_MessageControl, P_7_AB_1_PropertyName, VbLet, V_7_AB_1_Text
        End If
    End If
Else
    If IsWindow(P_7_AB_1_MessageHWnd) Then
        If IsWindowVisible(P_7_AB_1_MessageHWnd) Then
            V_7_AB_1_Res = SetWindowText(P_7_AB_1_MessageHWnd, V_7_AB_1_Text)
            If V_7_AB_1_Res = 0 Then
                V_7_AB_1_ErrorNumber = Err.LastDllError
                V_7_AB_1_ErrorText = F_7_AB_1_GetSystemErrorMessageText(V_7_AB_1_ErrorNumber)
                MsgBox "An error occurred with SetWindowText: " & vbCrLf & _
                       "System Error: " & CStr(V_7_AB_1_ErrorNumber) & vbCrLf & _
                       "Description:  " & V_7_AB_1_ErrorText
            End If
        End If
    End If
End If

End Sub


Private Function F_7_AB_1_GetSystemErrorMessageText(A_7_AB_1_ErrorNumber As Long) As String
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' F_7_AB_1_GetSystemErrorMessageText
'
' This function gets the system error message text that corresponds to the error code returned by the
' GetLastError API function or the Err.LastDllError property. It may be used ONLY for these error codes. These are NOT the error
' numbers returned by Err.Number (for these errors, use Err.Description to get the description of the message).
' The error number MUST be the value returned by GetLastError or Err.LastDLLError.
'
' In general, you should use Err.LastDllError rather than GetLastError because under some circumstances the value of
' GetLastError will be reset to 0 before the value is returned to VB. Err.LastDllError will always reliably return
' the last error number raised in a DLL.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

Dim V_7_AB_1_ErrorText As String
Dim V_7_AB_1_TextLen As Long
Dim V_7_AB_1_FormatMessageResult As Long
Dim V_7_AB_1_LangID As Long

' an error value of 0 indicates no error. Return an empty string rather than the
' default Windows system message text.
If A_7_AB_1_ErrorNumber = 0& Then
    F_7_AB_1_GetSystemErrorMessageText = ""
    Exit Function
End If



V_7_AB_1_LangID = 0&
V_7_AB_1_ErrorText = String$(C_7_AB_1_FORMAT_MESSAGE_TEXT_LEN, " ")
V_7_AB_1_TextLen = Len(V_7_AB_1_ErrorText)
On Error Resume Next
V_7_AB_1_FormatMessageResult = 0&

V_7_AB_1_FormatMessageResult = FormatMessage( _
                dwFlags:=FORMAT_MESSAGE_FROM_SYSTEM Or FORMAT_MESSAGE_IGNORE_INSERTS, _
                lpSource:=0&, _
                dwMessageId:=A_7_AB_1_ErrorNumber, _
                 dwLanguageId:=0&, _
                lpBuffer:=V_7_AB_1_ErrorText, _
                nSize:=Len(V_7_AB_1_ErrorText), _
                Arguments:=0&)
   
   
On Error GoTo 0
If V_7_AB_1_FormatMessageResult > 0 Then
    ' FormatMessage returned some text. Take the left V_7_AB_1_FormatMessageResult characters
    ' and return that text.
    V_7_AB_1_ErrorText = Left$(V_7_AB_1_ErrorText, V_7_AB_1_FormatMessageResult)
    F_7_AB_1_GetSystemErrorMessageText = V_7_AB_1_ErrorText
Else
    ' Format message didn't return any text. there is no text description for the specified error.
    F_7_AB_1_GetSystemErrorMessageText = "NO ERROR DESCRIPTION AVAILABLE"
End If


End Function

Private Sub S_7_AB_1_KillTheTimer()
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' S_7_AB_1_KillTheTimer
' This function kills the message interval timer.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
    Dim V_7_AB_1_Res As Long
    Dim V_7_AB_1_ErrorNumber As Long
    Dim V_7_AB_1_ErrorText As String
    
    ''''''''''''''''''''''''''''''''''''''''''''''''''''''
    ' Since we called SetTimer without an HWnd or idEvent (in case the
    ' user closed the form while the shell'd program was still running)
    ' we pass in 0& as the HWnd and the return code of SetTimer as the
    ' nIDEvent parameter to KillTimer. This will prevent an error
    ' if the form was closed while the Shell'd program was still
    ' running.
    ''''''''''''''''''''''''''''''''''''''''''''''''''''''
    V_7_AB_1_Res = KillTimer(hwnd:=0&, nIDEvent:=P_7_AB_1_TimerID)
    If V_7_AB_1_Res = 0 Then
        V_7_AB_1_ErrorNumber = Err.LastDllError
        V_7_AB_1_ErrorText = F_7_AB_1_GetSystemErrorMessageText(A_7_AB_1_ErrorNumber:=V_7_AB_1_ErrorNumber)
        MsgBox "An error occurred with KillTimer." & vbCrLf & _
            "System Error Number: " & CStr(V_7_AB_1_ErrorNumber) & vbCrLf & _
            "Description:         " & V_7_AB_1_ErrorText
    End If
End Sub


Private Function F_7_AB_1_GetHWndOfFormByControlVB6(A_7_AB_1_FormControl As VB.Control) As Long
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' F_7_AB_1_GetHWndOfFormByControlVB6
' This function returns the HWnd of the parent form of A_7_AB_1_FormControl.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
    Dim V_7_AB_1_Obj As Object
    Dim V_7_AB_1_Done As Boolean
    Set V_7_AB_1_Obj = A_7_AB_1_FormControl
    Do Until V_7_AB_1_Done
        Set V_7_AB_1_Obj = V_7_AB_1_Obj.Parent
        If Not TypeOf V_7_AB_1_Obj Is VB.Control Then
            V_7_AB_1_Done = True
            F_7_AB_1_GetHWndOfFormByControlVB6 = V_7_AB_1_Obj.hwnd
        End If
    Loop
End Function

Private Function F_7_AB_1_GetHWndOfFormByHWndVB6(A_7_AB_1_HWnd As Long) As Long
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' F_7_AB_1_GetHWndOfFormByHWndVB6
' This function returns the HWnd of the parent form of A_7_AB_1_HWnd.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

''''''''''''''''''''''''''''''''''''''''''
' ensure we have a valid window
''''''''''''''''''''''''''''''''''''''''''
If IsWindow(hwnd:=A_7_AB_1_HWnd) = 0 Then
    MsgBox "The HWnd in A_7_AB_1_HWnd to procedure F_7_AB_1_GetHWndOfFormByHWndVB6 is not a valid window."
    Exit Function
End If
''''''''''''''''''''''''''''''''''''''''''
' ensure we the window is visible
''''''''''''''''''''''''''''''''''''''''''
If IsWindowVisible(hwnd:=A_7_AB_1_HWnd) = 0 Then
    MsgBox "The HWnd in A_7_AB_1_HWnd to procedure F_7_AB_1_GetHWndOfFormByHWndVB6 is not a visible window."
    Exit Function
End If

F_7_AB_1_GetHWndOfFormByHWndVB6 = GetAncestor(A_7_AB_1_HWnd, GA_ROOT)



End Function

